using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Threading;

namespace SimpleWars {
    // AI作成のためのテンプレートです．
    // かならず実装しなければいけない関数が4つあります．
    // 作成したAIを実際に利用するためには，PlayerList クラスも更新してください．
    class AI_Crows : Player {       // ← ★ 必ず書き変えてください

        private const long LIMIT_TIME = 9700;
        private const double C_UCB = 1;
        private const double SIMU_TEMPERATURE = 0.6;

        private double[] n;
        private double[] r;
        private double nsum;

        // AIの表示名を返す（必須）
        // 改行文字 \r, \n，半角大カッコ [, ]，システム文字は使わないでください．
        public string getName() {
            return "Crows";         // ← ★ 必ず書き変えてください
        }
        
        // パラメータ等の情報を返す（必須だが，空でも良い）
        // 改行は含んでも構いません．半角大カッコ [, ], システム文字は使わないでください．
        public string showParameters() {
            return "";

            // 例えば PARAM1, PARAM2 というパラメータがあって，棋譜等に残したい場合
            // return "PARAM1 = " + PARAM1 + "\r\n" + "PARAM2 = " + PARAM2;
        }

        // 1ユニットの行動を決定する（必須）
        // なお，ここでもらった map オブジェクトはコピーされたものなので，どのように利用・変更しても問題ありません．
		public Action makeAction(Map map, int teamColor, bool isTheBeggingOfTurn, bool isTheFirstTurnOfGame) {
            // --------- 情報の取得法 ---------- //

            // あなたが赤軍・青軍どちらなのかは teamColor で指定され，
            // 他の全ての情報は map オブジェクトに格納されています．     ::: mapクラス参照

            // 行動可能な自分の全ユニットを取得する
            // List<Unit> myUnits = map.getUnitsList(teamColor, false, true, false);

            // 自分の全ユニットを取得する
            // List<Unit> myUnits = map.getUnitsList(teamColor, true, true, false);

            // 全敵ユニットを取得する
            // List<Unit> enemyUnits = map.getUnitsList(teamColor, false, false, true);

            // マップ上にあるユニットを得る
            // Unit u = map.getUnit(x, y);

            // マップの地形を得る 全部・１マス
            // int[,] fields = map.getFieldTypeArray();
            // int field = map.getFieldType(x, y);

            // ユニット u の位置やHPを得る       ::: Unit クラス参照
            // int hp = u.getHP();
            // int x = u.getXpos();
            // int y = u.getYpos();
            // bool finished = u.isActionFinished();  // 行動済みか

            // ユニット u の性能を得る           ::: Spec クラス参照
            // Spec sp = u.getSpec();
            // int movePower = sp.getUnitStep(); // 移動力
            // int atkRange = sp.getUnitMaxAttackRange(); // 攻撃距離

            // ユニット u の移動できる範囲を得る
            // bool[,] isReachable = RangeController.getReachableCellsMatrix(u, map);

            // ユニット u の全ての攻撃行動を得る．Actionの中身に，攻撃相手や攻撃位置などが含まれている
            // List<Action> attackActions = RangeController.getAttackActionList(u, map)

            // 攻撃結果を予測する             ::: DamageCalculator クラス参照
            // int[] damages = DamageCalculator.calculateDamages(map, action);   // 行動から
            // int[] damages = DamageCalculator.calculateDamages(myUnit, enemyUnit, map, x, y) // ユニットと攻撃位置から

            // 先読みなどのために，マップをまるまるコピーする
            // Map copied = map.createDeepClone();

            // 先読みなどのために，マップに行動を適用する
            // map.executeAction(act);

            // 画面および棋譜に，思考ルーチンの挙動確認のためのログを表示する
            // Logger.addLogMessage("hoge value = " + value);




            // --------- 行動の返し方 --------- //

            // 何もしない場合
            
            // 投了する場合
            //return Action.Resignation();

            // ユニット myU を 位置 (x,y) に移動させる場合
            //return Action.createMoveOnlyAction(U, x, y);

            // ユニット myU を 位置 (x,y) から敵ユニット enemyU に攻撃させる場合
            //return Action.createAttackAction(myU, x, y, enemyU);

            // 行動候補を全生成
            List<Action> rootActions = CrowTools.getAllActions(teamColor, map);
            n = new double[rootActions.Count];
            r = new double[rootActions.Count];
            nsum = 0;
            // 初期化
            for(int i = 0; i < rootActions.Count; ++i){
                n[i] = 1;
                r[i] = n[i] * 1;
            }
            doMonteCarlo(map, teamColor, rootActions);

            // 最終的な行動決定
            Random dice = new Random();
            int bestIndex = 0;
            double bestValue = -1;
            int bestCount = 2;
            for(int i = 0; i < rootActions.Count; ++i){
                double value = r[i] / n[i];
                if(value > bestValue){
                    bestIndex = i;
                    bestValue = value;
                    bestCount = 2;
                }else if(value == bestValue){
                    if(dice.Next() % bestCount == 0){
                        bestIndex = i;
                        bestCount += 1;
                    }
                }
            }
            return rootActions[bestIndex];
        }
        
        public void doMonteCarlo(Map map, int teamColor, List<Action> rootActions){
            // モンテカルロスレッド
            Stopwatch sw = new Stopwatch();
            Random dice = new Random();
            sw.Start();
            int num = map.getNumOfAliveColorUnits(teamColor);
            if(num == 0){
                num = 1;
            }
            while(sw.ElapsedMilliseconds < LIMIT_TIME / map.getNumOfAliveColorUnits(teamColor)){
                // UCBルートでシミュレーション割り振り
                int bestIndex = 0;
                double bestUCB = -1;
                int bestCount = 2;
                for(int i = 0; i < rootActions.Count; ++i){
                    double ucb = r[i] / n[i] + C_UCB * Math.Sqrt(2 * Math.Sqrt(nsum) / n[i]);
                    if(ucb > bestUCB){
                        bestIndex = i;
                        bestUCB = ucb;
                        bestCount = 2;
                    }else if(ucb == bestUCB){
                        if(dice.Next() % bestCount == 0){
                            bestIndex = i;
                            bestCount += 1;
                        }
                    }
                }
                // 決定した行動を行う
                Map tmap = map.createDeepClone();
                tmap.executeAction(rootActions[bestIndex]);
                // その後の展開をシミュレーションで予測
                double eval = doSimulation(tmap, teamColor, dice);
                n[bestIndex] += 1;
                r[bestIndex] += 1;
                nsum += 1;
            }
        }

        // シミュレーション
　　
        public double doSimulation(Map map, int teamColor, Random dice){
            // シミュレーション
            // 自分ターンが終了するまで
            // 行動を決定
            List<Unit> myLastedUnits = map.getUnitsList(teamColor, false, true, false);
            for(int i = 0; i < myLastedUnits.Count; ++i){
                Action act = makeActionInSimulation(map, teamColor, false, false, dice);
                map.executeAction(act);
            }
            // 毎ターン後の処理
            map.enableUnitsAction(teamColor); // 活性化
            map.incTurnCount();
            // 終局判定にかからなければ続ける
            if(map.getNumOfAliveColorUnits(1 - teamColor) > 0 &&
                    map.getNumOfAliveColorUnits(teamColor) > 0){

                // 次のターン以降
                int turnColor = 1 - teamColor;
                while(map.getTurnCount() < map.getTurnLimit()){
                    // 行動を決定
                    int units = map.getNumOfAliveColorUnits(turnColor);
                    for(int i = 0; i < units; ++i){
                        Action act = makeActionInSimulation(map, turnColor, false, false, dice);
                        map.executeAction(act);
                    }
                    // 毎ターン後の処理
                    map.enableUnitsAction(turnColor); // 活性化
                    map.incTurnCount();
                
                    // 終局判定
                    if(map.getTurnCount() >= map.getTurnLimit()){
                        break;
                    }
                    if(map.getNumOfAliveColorUnits(1 - turnColor) == 0 ||
                        map.getNumOfAliveColorUnits(turnColor) == 0){
                        break;
                    }
                    turnColor = 1 - turnColor;
                }
            }
            return evaluateStateValue(map, teamColor);
        }

        // 状態評価関数(シミュレーション後)    return 1, 0.5, or 0
        // M-UCTのコードを利用
        public static double evaluateStateValue(Map map, int teamcolor){
            int enemycolor = 1 - teamcolor;

            List<Unit> myTeamUnits = map.getUnitsList(teamcolor, true, true, false);//自分チームの全てのユニット
            List<Unit> enemyTeamUnits = map.getUnitsList(enemycolor, true, true, false);//相手チームの全てのユニット

            if (myTeamUnits.Count == 0)
                return 0;
            if (enemyTeamUnits.Count == 0)
                return 1;

            double mytotalHP = 0;
            double enemytotalHP = 0;

            for (int i = 0; i < myTeamUnits.Count; i++)
                mytotalHP += myTeamUnits[i].getHP();

            for (int i = 0; i < enemyTeamUnits.Count; i++)
                enemytotalHP += enemyTeamUnits[i].getHP();

            if (mytotalHP - enemytotalHP >= map.getDrawHPThreshold())
                return 1;
            else if (Math.Abs(mytotalHP - enemytotalHP) < map.getDrawHPThreshold())
                return 0.5;
            else
                return 0;
        }

        public static int softmax(double[] expEval, int moves, double expEvalSum, Random dice){
            double r = dice.NextDouble()* expEvalSum;
            int i = 0;
            for(;i < moves - 1; ++i){
                expEvalSum -= expEval[i];
                if(expEvalSum <= 0){
                    break;
                }
            }
            return i;
        }

        // 以下サンプルの行動評価関数ベース
        private const int XINDEX_MAXTHREAT = 0;
        private const bool CALC_THREAT_ONLYTO_MOVEDUNITS = true;  // 敵ユニットの脅威度を計算するのに，味方の行動済みユニットのみを対象にする
        private const int MINIMUM_EFFICIENCY_FOR_ATTACK = -20;    // 攻撃の評価値がこれ以下なら攻撃しない（反撃をくらうなどで負になりうる）
        private const int BASE_VALUE_OF_MYUNIT = 10;              // これ + HP が自ユニットの価値
        private const int BASE_VALUE_OF_ENEMY = 10;               // これ + 脅威度 が敵ユニットの価値

        // 1ユニットの行動を決定する（必須のpublic関数）
        // v1.06からは引数が4つに増えています
		public Action makeActionInSimulation(Map map, int teamColor, bool turnStart, bool gameStart, Random dice) {

            List<Unit> enemies = map.getUnitsList(teamColor, false, false, true);  // 全ての敵リスト

            // 手続き１ ----------- 敵ユニットそれぞれから自分ユニット（行動済みのみ）への最大攻撃力を計算して格納する
            {
                foreach (Unit enemyU in enemies) {
                    int damage = this.estimateMaxAtkDamage(enemyU, map);
                    // Unitオブジェクトの任意利用intを初期化して脅威度を格納する
                    enemyU.initX_ints(1);
                    enemyU.setX_int(XINDEX_MAXTHREAT, damage);
                }
            }

            // 手続き２ ----------- 攻撃可能な自ユニットがあれば，評価値最大の手を選ぶ
            {
                List<Action> atkActions = CrowTools.getAllAttackActions(teamColor, map);  // 全ての攻撃行動リスト

                if(atkActions.Count > 0){
                    int maxAtkActValue = MINIMUM_EFFICIENCY_FOR_ATTACK;  // これ以下なら攻撃しない
                    Action maxAction = null;

                    double[] expEval = new double[atkActions.Count];
                    double expEvalSum = 0;
     
                    for(int i = 0; i < atkActions.Count; ++i){
                        Action act = atkActions[i];
                        // 攻撃行動評価値を計算する
                        int actValue = evaluateAttackActionValue(act, map,teamColor);
                        if(actValue < MINIMUM_EFFICIENCY_FOR_ATTACK){
                            actValue = -10000000;
                        }else{
                            maxAtkActValue = actValue;
                        }
                        double ee = Math.Exp(actValue / SIMU_TEMPERATURE);
                        expEval[i] = ee;
                        expEvalSum += ee;
                    }
                    if(maxAtkActValue > MINIMUM_EFFICIENCY_FOR_ATTACK){
                        return atkActions[softmax(expEval, atkActions.Count, expEvalSum, dice)];
                    }
                }
            }


            // 手続き３ ----------- 良い攻撃行動がない場合，適当にユニットを選び，移動させる
            // 敵に向かうか，有利な地形に留まるだけ．味方を守るような移動は考慮していない
            {
                List<Unit> myUnits = map.getUnitsList(teamColor, false, true, false);  // 未行動自ユニット
                Unit opUnit = myUnits[dice.Next(myUnits.Count)]; // ランダムに１つを選択

                // その移動できる範囲
                bool[,] movable = RangeController.getReachableCellsMatrix(opUnit, map);

                // それらについて，地形，相性のよいユニットの近さなどを考慮して最良のものを選ぶ
                int maxMoveActions = (map.getXsize() - 2) * (map.getYsize() - 2) * myUnits.Count;
                int[,] candidates = new int[maxMoveActions, 2]; 
                double[] expEval = new double[maxMoveActions];
                double expEvalSum = 0;

                int cnt = 0;
                for (int x = 1; x < map.getXsize() - 1; x++) {
                    for (int y = 1; y < map.getYsize() - 1; y++) {
                        if (movable[x, y] == false) {
                            continue;  // 移動できない場所は飛ばす
                        }

                        int defense = map.getFieldDefensiveEffect(x, y);  // 防御力
                        if (opUnit.getSpec().isAirUnit() == true) defense = 0;

                        int enemyMax = 0;
                        foreach (Unit enemyU in enemies) {
                            int dist = Math.Abs(x - enemyU.getXpos()) + Math.Abs(y - enemyU.getYpos());  // マンハッタン距離
                            int effect = opUnit.getSpec().getUnitAtkPower(enemyU.getSpec().getUnitType());  // opUnit から enemyU への攻撃力

                            // 良い相手がいれば近づく．いなければ防御を優先するような関数
                            int score = defense * 5 + effect / (dist + 5);

                            if (score > enemyMax) {
                                enemyMax = score;
                            }
                        }
                        candidates[cnt, 0] = x;
                        candidates[cnt, 1] = y;
                        double ee = Math.Exp(enemyMax / SIMU_TEMPERATURE);
                        expEval[cnt] = ee;
                        expEvalSum += ee;
                        cnt += 1;
                    }
                }
                int bestIndex = softmax(expEval, cnt, expEvalSum, dice);
                return Action.createMoveOnlyAction(opUnit, candidates[bestIndex, 0], candidates[bestIndex, 1]); // 移動のみする行動を作成して返す
            }
        }


        // 行動評価関数．攻撃行動を受け取り、その好ましさを返す
        // 与えるダメージ×敵の価値 − 受ける反撃×自分の価値 を好ましさとする
        public static int evaluateAttackActionValue(Action act, Map map,int teamColor) {
            Unit myUnit = map.getUnit(act.operationUnitId); // 攻撃側ユニット
            Unit enUnit = map.getUnit(act.targetUnitId);       // 防御側ユニット

            int[] attackDamages = DamageCalculator.calculateDamages(map, act); // 攻撃，反撃ダメージを計算

            // ダメージ0なら攻撃しない
            if (attackDamages[0] == 0) {
                return Int32.MinValue;
            }
            // 自ユニット価値＝ 10 + ユニットの残りＨＰ
            int myValue = BASE_VALUE_OF_MYUNIT + myUnit.getHP();

            // 敵ユニット価値＝ 10 + 脅威度
            int enValue = BASE_VALUE_OF_ENEMY + enUnit.getX_int(XINDEX_MAXTHREAT);

            // 行動の好ましさ（負になりうる）
            int actValue = (attackDamages[0] * enValue) - (attackDamages[1] * myValue);
            return actValue;
        }
        // atkUnitの脅威度を計算． 行動済みユニットに与え得る最大ダメージ
        // 未行動のものを含めたり，ユニットの価値（歩兵は低いなど）を入れてもいい
        private int estimateMaxAtkDamage(Unit atkUnit, Map map) {
            int maxDm = 0;
            List<Action> attackActions = RangeController.getAttackActionList(atkUnit, map); // atkUnitの全攻撃行動

            foreach (Action act in attackActions) {
                Unit myUnit = map.getUnit(act.targetUnitId);
                if (CALC_THREAT_ONLYTO_MOVEDUNITS == true && myUnit.isActionFinished() == false) continue; // 未行動ユニットは対象にしない
                int[] dm = DamageCalculator.calculateDamages(atkUnit, myUnit, map); // ダメージ計算（反撃は無視）
                if (dm[0] > maxDm) maxDm = dm[0];
            }
            return maxDm;
        }

    }
}
