﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace SimpleWars {
    
    /// <summary>
    /// 読むのが面倒なメソッド群のまとめ
    /// </summary>
    static class SAT_FUNC {

       

        //teamColorの反対の色を返す
        public static int REVERSE_COLOR(int teamColor) {
            if (teamColor == Consts.BLUE_TEAM) return Consts.RED_TEAM;
            return Consts.BLUE_TEAM;
        }

        
        /***************
         ***
         ***  自軍の一部ユニット切り抜きに使うメソッド群
         ***　　　↓↓↓
         **************/

        //自ユニット（行動可能）のMap上全ユニットの中心位置に近い4体のIDを順にキューに詰めて返す
        public static Queue<int> makeUnitQ_forward(Map map, int teamColor) {
            //Map状の全ユニットの中心位置．{x, y}
            int[] gravity_xy = CALC_AVERAGED_2TEAM_GRAVITY_POINT(map);
            int gra_x = gravity_xy[0];
            int gra_y = gravity_xy[1];

            List<Unit> units = map.getUnitsList(teamColor, false, true, false);

            //※記法がややこしいが，要するにユニットをMapの重心座標に近い順に整頓してるだけ
            units.Sort((a, b) => Math.Abs(a.getXpos() - gra_x) + Math.Abs(a.getYpos() - gra_y) -
                                                    Math.Abs(b.getXpos() - gra_x) - Math.Abs(b.getYpos() - gra_y));

            Queue<int> unitQue_forward = new Queue<int>();
            int maxLen = Math.Min(4, units.Count);
            for (int i = 0; i < maxLen; i++) {
                unitQue_forward.Enqueue(units.ElementAt(i).getID());
            }
            return unitQue_forward;
        }

        //自ユニット（行動可能）のMap上全ユニットの中心位置に遠い4体を "逆順" にキューに詰めて返す
        public static Queue<int> makeUnitQ_reverse(Map map, int teamColor) {
            //Map状の全ユニットの中心位置．{x, y}
            int[] gravity_xy = CALC_AVERAGED_2TEAM_GRAVITY_POINT(map);
            int gra_x = gravity_xy[0];
            int gra_y = gravity_xy[1];

            List<Unit> units = map.getUnitsList(teamColor, false, true, false);

            //※記法がややこしいが，要するにユニットをMapの重心座標に近い順に整頓してるだけ
            units.Sort((a, b) => Math.Abs(a.getXpos() - gra_x) + Math.Abs(a.getYpos() - gra_y) -
                                                    Math.Abs(b.getXpos() - gra_x) - Math.Abs(b.getYpos() - gra_y));
            units.Reverse();
            Queue<int> unitQue_backward = new Queue<int>();
            int maxLen = Math.Min(4, units.Count);
            for (int i = 0; i < maxLen; i++) {
                unitQue_backward.Enqueue(units.ElementAt(units.Count - maxLen + i).getID());
            }
            return unitQue_backward;
        }

        //マップ上の全ユニット位置のおよその中心座標の導出
        public static int[] CALC_AVERAGED_2TEAM_GRAVITY_POINT(Map map) {
            int[] grav_xy_R = new int[2]; //赤チームのユニット重心座標 {X, Y}
            int[] grav_xy_B = new int[2]; //青チームの　〃
            int total_R = 0;
            int total_B = 0;
            Unit[] uniList = map.getUnits();
            foreach (Unit u in uniList) {
                if (u == null || u.isDead()) { continue; }
                if (u.getTeamColor() == Consts.RED_TEAM) {
                    grav_xy_R[0] += u.getXpos();
                    grav_xy_R[1] += u.getYpos();
                    total_R++;
                } else {
                    grav_xy_B[0] += u.getXpos();
                    grav_xy_B[1] += u.getYpos();
                    total_B++;
                }
            }
            if (total_R == 0) {
                grav_xy_R = new int[2] { map.getXsize() / 2, map.getYsize() / 2 };
            } else {
                grav_xy_R[0] /= total_R;
                grav_xy_R[1] /= total_R;
            }
            if (total_B == 0) {
                grav_xy_B = new int[2] { map.getXsize() / 2, map.getYsize() / 2 };
            } else {
                grav_xy_B[0] /= total_B;
                grav_xy_B[1] /= total_B;
            }
            grav_xy_R[0] = (grav_xy_R[0] + grav_xy_B[0]) / 2;
            grav_xy_R[1] = (grav_xy_R[1] + grav_xy_B[1]) / 2;
            return grav_xy_R;
        }


        /***************
         ***
         ***  合法手（のうち探索に使うもの）の抜き出しに使うメソッド群
         ***　　　↓↓↓
         **************/

        public static List<Action> suggestMoves(Map map, Unit unit_to_move, int teamColor) {
            List<Action> actList = new List<Action>();
            //## ATTACK ##
            actList.AddRange(suggest1AtkActsforEachEnemy(map, unit_to_move, teamColor));
            //## MOVE ## 
            actList.AddRange(suggestMeaningfulMoveActs(map, unit_to_move, teamColor));            
            return actList;
        }

        //攻撃対象となる敵ユニットがそれぞれ異なる攻撃行動の獲得
        private static List<Action> suggest1AtkActsforEachEnemy(Map map, Unit unit_to_move, int teamColor) {
            List<Action> eachEneAtkActs = new List<Action>();
            List<Action> acts = AiTools.getAllAttackActions(teamColor, map);
            
            HashSet<int> attackedIDs = new HashSet<int>();

            foreach(Action act in acts){                
                if (act.operationUnitId != unit_to_move.getID()) { continue; }//引数のユニットの行動じゃない
                if(attackedIDs.Contains(act.targetUnitId)) {       continue; }// 既に攻撃を考慮した事のある敵ユニット
                if (!isEffective(map.getUnit(act.operationUnitId), map.getUnit(act.targetUnitId))) { continue; }//効果なし攻撃
                eachEneAtkActs.Add(act);
                attackedIDs.Add(act.targetUnitId);
            }
            return eachEneAtkActs;
        }

        //敵からの制空権パターンがそれぞれ異なる位置への移動行動の獲得
        private static List<Action> suggestMeaningfulMoveActs(Map map, Unit unit_to_move, int teamColor) {
            List<Action> meaningfulActs = new List<Action>();
            
            List<Action> acts = AiTools.getAllMoveActions(teamColor, map);
            HashSet<int> movedAreaFlags = new HashSet<int>();
            
            foreach (Action act in acts) {
                if (act.operationUnitId != unit_to_move.getID()) { continue; }//引数のユニットの行動じゃない
                int destX = act.destinationXpos;
                int destY = act.destinationYpos;
                int masteryBitFlag = SAT_RangeController.enemyMastery[destX, destY];
                if (movedAreaFlags.Contains(masteryBitFlag)) { continue; }// 既に攻撃を考慮した事のある敵ユニット
                meaningfulActs.Add(act);
                movedAreaFlags.Add(masteryBitFlag);
            }
            return meaningfulActs;
        }




        /***************
         ***
         ***  局面評価に使うメソッド群
         ***　　　↓↓↓
         **************/


        //味方から敵に一斉に一方的攻撃をしてみた場合の敵残りHP残量を計算して返す
        //※　もちろん実際局面からの誤差は小さくないが，速めに計算処理できる．
        public static int evalState_withAtkPrediction(Map map, int teamColor) {
            List<Unit> myUnits = map.getUnitsList(teamColor, true, true, false);
            List<Unit> eneUnits = map.getUnitsList(teamColor, false, false, true);
                        
            int eneHpSum = 0;
            //初期HPの保存
            foreach (Unit eneU in eneUnits) {
                if (eneU.getHP() <= 0) { continue; }//死んだユニットは，どんなバグ出すか解らんので触らない！
                eneHpSum += eneU.getHP();
            }
            
            int friendHpSum = 0;
            //attack "Friend -> Enemy"
            foreach (Unit myU in myUnits) {
                if (myU.getHP() <= 0) { continue; }
                friendHpSum += myU.getHP();//HP保存

                int benefit = 0;
                foreach (Unit eneU in eneUnits) {
                    if (eneU.getHP() <= 0 ) { continue; }
                    if (!IN_RANGE_ESTIMATE(map, myU, eneU)) { continue; }
                    if (!SAT_FUNC.isEffective(myU, eneU)) { continue; }

                    //##このCalcDamages()はダメージが敵HPを超えないよう自動的に調整してくれてる事に留意##
                    int[] dam_GiveBack = DamageCalculator.calculateDamages(myU.getSpec(), myU.getHP(),
                                eneU.getSpec(), eneU.getHP(),
                                map.getFieldDefensiveEffect(myU.getXpos(), myU.getYpos()),
                                map.getFieldDefensiveEffect(eneU.getXpos(), eneU.getYpos())
                                );
                    benefit = Math.Max(benefit, dam_GiveBack[0] - dam_GiveBack[1]);
                }
                eneHpSum -= benefit;
            }


            //Calc. HP-Diff. HP情報を復元する都合上，下記のEval＿Map_HPメソッドを流用してはいけない．
            int HPdiff = friendHpSum - eneHpSum;
            return HPdiff;
        }




        /// <summary>
        /// Estimate the attackUnit can attack to the defendUnit.
        /// </summary>
        private static bool IN_RANGE_ESTIMATE(Map map, Unit attackU,
                                                Unit defendU) {
            //遠距離ユニットは判定が簡単なので，精密近似/ラフ近似の別に関わらず，さっさと処理して値を返してしまう．
            if (!attackU.getSpec().isDirectAttackType()) {
                int distCannon = Math.Abs(attackU.getXpos() - defendU.getXpos()) 
                                    + Math.Abs(attackU.getYpos() - defendU.getYpos());
                return (2 <= distCannon && distCannon <= 3);
            }

            //ラフに推測して返す．(飛行系ユニットか歩兵)
            if (attackU.getSpec().isAirUnit() || attackU.getSpec().getUnitStep() == 3) {
                int dist = Math.Abs(attackU.getXpos() - defendU.getXpos())
                                    + Math.Abs(attackU.getYpos() - defendU.getYpos());
                return (dist <= attackU.getSpec().getUnitStep() + 1);
            }
            //ラフに推測して返す．(戦車系ユニット)
            int remStep = SAT_RangeController.rangeMap[attackU.getXpos(), attackU.getYpos()][defendU.getXpos(), defendU.getYpos()];
            return remStep >= 0;
        }

        //攻防ユニット間にダメージが通る相性であるか．　ダメージ係数が非ゼロ．
        public static bool isEffective(Unit operationUnit, Unit targetUnit) {
            if (operationUnit.getSpec().getUnitAtkPower(targetUnit.getTypeOfUnit()) != 0) { return true; }
            return false;
        }     
        
     
    }
}