﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Diagnostics;
using SimpleWars;
using System.Threading.Tasks;
using System.Threading;


namespace SimpleWars
{
    class Misirlou_v3 : Player
    {
        //Main parameters
        private static int max_sim = 30000; //Maximum of simulation number
        private static int search_depth = 0;//depth you can search
        private static int max_child = 0;
        private const double z = 2;//control UCB formula
        private static double oneunit_sim;

        //Fixed pattern order control parameters
        private static List<Unit> order = new List<Unit>(); //This controls the simulation order

        //Time control parameters
        private static long Lim_time = 9900;//Limited time for every turn
        private static long Lef_time = 0;
        Stopwatch timer = new Stopwatch();//Creating timer

        //random number creator
        Random rand = new Random();

        #region player界面所需参数
        //返回AI名
        public string getName()
        {
            return "Misirlou_v3";
        }

        // 返回参数信息
        public string showParameters()
        {
            return "";
        }
        #endregion

        //Main logic module
        public Action makeAction(Map map, int teamColor, bool turnBegin, bool gameBegin)
        {
            timer.Start();
            int now_sim = 0; //Simulations number for now
            int movablenum = map.getUnitsList(teamColor, false, true, false).Count;
            nodes rootnode = new nodes();
            rootnode.map = map.createDeepClone();
            rootnode.color = teamColor;//队伍颜色参数在其他类中得到
            List<Action> AttackActionList = mSpecTools.getAllAttackActions(rootnode.color, rootnode.map);
            List<Action> MoveActionList = mSpecTools.getAllMoveActions(rootnode.color, rootnode.map);

            if (turnBegin == true)
            {
                oneunit_sim = max_sim / map.getUnitsList(teamColor, false, true, false).Count;
                Lef_time = Lim_time;
            }

            for (int a = 1; a <= max_sim; a++)
            {
                if (timer.ElapsedMilliseconds > (Lef_time / movablenum))
                {
                    break;
                }
                progressiveWiding(rootnode, a, AttackActionList, MoveActionList);
                UCT(rootnode);
                now_sim++;
            }

            Action bestmove = best_move(rootnode);
            timer.Stop();

            Logger.addLogMessage("Numbers of simulation: " + now_sim.ToString() + "\r\n", teamColor);
            Logger.addLogMessage("bestmove win number: " + rootnode.child[max_child].win_Num.ToString() + "\r\n", teamColor);
            Logger.addLogMessage("bestmove sim number: " + rootnode.child[max_child].sim_Num.ToString() + "\r\n", teamColor);
            Logger.addLogMessage("Max win ratio: " + rootnode.child[max_child].win_ratio.ToString() + "\r\n", teamColor);
            Logger.addLogMessage("Searching Depth = " + search_depth.ToString() + "\r\n", teamColor);
            Logger.addLogMessage("Action Type: " + bestmove.actionType.ToString() + "\r\n", teamColor);
            Logger.addLogMessage("Best Action: From " + bestmove.fromXpos.ToString() + "," + bestmove.fromYpos.ToString() + "\r\n" +
                                 "             To   " + bestmove.destinationXpos.ToString() + "," + bestmove.destinationYpos.ToString() + "\r\n\r\n", teamColor);
            Lef_time -= timer.ElapsedMilliseconds;
            timer.Reset();
            return bestmove;
        }

        //得到最大UCB值节点，并通过常数控制保证没被模拟过的节点仍会被选择到
        private Action best_move(nodes rootnode)
        {
            double maxwin_ratio = -1;
            nodes bestnode = new nodes();
            rand = new Random();

            for (int i = 0; i < rootnode.child.Count; i++)
            {
                if (rootnode.child[i].win_ratio > maxwin_ratio)
                {
                    if (rootnode.child[i].actions.actionType == 2)
                    {
                        if (stopsuicide(rootnode,rootnode.child[i]) == true)
                        {
                            continue;
                        }
                    }
                    maxwin_ratio = rootnode.child[i].win_ratio;
                    max_child = i;
                    bestnode = rootnode.child[i];
                }
            }

            if (maxwin_ratio <= 0.2)//进入函数判断阶段,此处可以控制阈值
            {
                bestnode.actions = MaxUnitdeterminate(rootnode);
            }

            return bestnode.actions;
        }

        //一次完整的UCT过程
        public void UCT(nodes rootnode)
        {
            int d = 0;//depth
            nodes[] oneway = new nodes[max_sim];


            //拓展根节点，根节点的子节点中存入actions

            oneway[d] = rootnode;

            //节点存在子节点的情况下则进行selection
            while (oneway[d].child.Count() > 0)
            {
                oneway[d + 1] = UCBselection(oneway[d]);
                d++;
            }
            //没有子节点存在则进行模拟（模拟过程包含扩展）
            simulate(oneway[d]);

            if (search_depth < d)//如果此次为最大深度 更新搜索深度
            {
                search_depth = d;
            }

            //backpropagation，更新oneway所有节点的模拟数，胜利数，UCB值
            for (int depth = d; depth > 0; depth--)//遍历所有父节点
            {
                if (oneway[depth].color == oneway[depth - 1].color)//父节点与子节点颜色相同
                {
                    oneway[depth - 1].sim_Num += oneway[depth].sim_Num;//更新模拟数
                    oneway[depth - 1].win_Num += oneway[depth].win_Num;//更新胜利数
                    oneway[depth].win_ratio = oneway[depth].win_Num / oneway[depth].sim_Num;
                    oneway[depth].ucb = oneway[depth].win_Num / oneway[depth].sim_Num + Math.Sqrt(z * Math.Log(oneway[depth - 1].sim_Num) / oneway[depth].sim_Num);//更新末端节点ucb
                }
                else//父节点与子节点颜色不同
                {
                    oneway[depth - 1].sim_Num += oneway[depth].sim_Num;//更新模拟数
                    oneway[depth].win_ratio = oneway[depth].win_Num / oneway[depth].sim_Num;
                    oneway[depth].ucb = (1 - oneway[depth].win_Num / oneway[depth].sim_Num) + Math.Sqrt(z * Math.Log(oneway[depth - 1].sim_Num) / oneway[depth].sim_Num);//计算末端ucb

                }
            }
        }

        //到达末端结点则进行一次终盘模拟
        private void simulate(nodes node)
        {
            Map map = node.map.createDeepClone();//末端节点的地图信息

            //叶子结点为最终节点
            if (evaluateMAP(node.map, 1) != -1)//即游戏已结束
            {
                if (evaluateMAP(node.map, node.color) == 1)
                {
                    node.sim_Num++;
                    node.win_Num++;
                }
                else
                {
                    node.sim_Num++;//模拟数+1
                }
            }
            else//游戏没结束则扩展节点
            {
                expand(node);//扩展末端节点,被扩展的子节点内只有Action

                int tempNodeNum = rand.Next(node.child.Count);//随机选择一个子节点
                int nowColor;
                Map map2 = node.child[tempNodeNum].map.createDeepClone();//创造末端子节点的地图

                if (map2.getUnitsList(node.color, false, true, false).Count == 0)//我方可行动单位为零，子节点行动都为对方
                {
                    map2.resetActionFinish();//重置动作
                    map2.incTurnCount();
                    nowColor = (node.child[tempNodeNum].color + 1) % 2;//颜色变为对方颜色
                }
                else
                {
                    nowColor = node.child[tempNodeNum].color;//下一步仍为己方
                }

                //循环进行模拟直到终局
                for (int k = map2.getTurnCount(); k < map2.getTurnLimit() && evaluateMAP(map2, 0) == -1; k++)
                {
                    doOneTurnAction(map2, nowColor++ % 2);
                }

                //更新末端节点胜利数与模拟数
                if (evaluateMAP(map2, node.color) == 1)//终局模拟返回为胜
                {
                    node.win_Num++;//末端子节点胜利数+1
                    node.sim_Num++;//末端子节点模拟数+1

                }
                else
                {
                    node.sim_Num++;
                }

            }
        }

        //UCB选择模块
        private nodes UCBselection(nodes node)
        {
            rand = new Random();
            double maxUcb = 0;
            int maxUcbChild = 0;
            int a = 0;
            nodes Maxnode;

            //遍历子节点
            for (a = 0; a < node.child.Count; a++)
            {
                nodes child = node.child[a];

                if (child.sim_Num == 0)
                {
                    child.ucb = 10;
                }
                if (child.ucb > maxUcb)
                {
                    maxUcb = child.ucb;
                    maxUcbChild = a;
                }
            }
            Maxnode = node.child[maxUcbChild];
            return Maxnode;
        }


        //扩展结点
        private void expand(nodes node)
        {
            Map map2 = node.map.createDeepClone();
            List<Action> MoveActionList;//移动指令序列
            List<Action> AttackActionList;//攻击指令序列
            int nexTeamCol;

            //列举可能行动之前检查行动队伍是否发生改变
            if (map2.getUnitsList(node.color, false, true, false).Count == 0)//我方可用单位为零
            {//下一手变为对方行动
                map2.resetActionFinish();//所有剩余单位变为可行动状态
                map2.incTurnCount();//回合数+1
                nexTeamCol = (node.color + 1) % 2;//蓝队变红队，红队变蓝队，可行动方交换
                MoveActionList = mSpecTools.getAllMoveActions((node.color + 1) % 2, map2);//移动指令变为对方可移动指令
                AttackActionList = mSpecTools.getAllAttackActions((node.color + 1) % 2, map2);//攻击指令变为对方可攻击指令
            }
            else//我方可用单位不为零
            {//下一手仍为我方行动
                nexTeamCol = node.color;//下一手仍然是自己的颜色
                MoveActionList = mSpecTools.getAllMoveActions(node.color, map2);//仍然是自己的移动行动指令
                AttackActionList = mSpecTools.getAllAttackActions(node.color, map2);//仍然是自己的攻击指令
            }

            for (int i = 0; i < AttackActionList.Count; i++)//对攻击指令进行遍历
            {
                node.child.Add(new nodes());//添加子节点
                node.child[i].actions = AttackActionList[i];//添加攻击指令进指令集
                node.child[i].map = map2.createDeepClone();//添加地图信息
                node.child[i].map.executeAction(AttackActionList[i]);
                node.child[i].color = nexTeamCol;//进入下一手、
            }
            for (int i = 0; i < MoveActionList.Count; i++)//对移动指令进行遍历
            {
                node.child.Add(new nodes());//添加子节点
                node.child[i + AttackActionList.Count].actions = MoveActionList[i];//添加移动指令进指令集
                node.child[i + AttackActionList.Count].map = map2.createDeepClone();//复制地图信息
                node.child[i + AttackActionList.Count].map.executeAction(MoveActionList[i]);
                node.child[i + AttackActionList.Count].color = nexTeamCol;//进入下一手
            }
        }

        #region 扩展一个节点
        private void expand1node(nodes node)
        {
            Map map2 = node.map.createDeepClone();
            List<Action> AttackActionList;
            List<Action> ActionList;

            int nexTeamCol;
            int aTmp;
            Random aRand = RandomT.GetThreadRandom();

            //列举可能行动之前检查行动队伍是否发生改变
            if (map2.getUnitsList(node.color, false, true, false).Count == 0)//我方可用单位为零
            {//下一手变为对方行动
                map2.resetActionFinish();//所有剩余单位变为可行动状态
                map2.incTurnCount();//回合数+1
                nexTeamCol = (node.color + 1) % 2;//蓝队变红队，红队变蓝队，可行动方交换
                ActionList = mSpecTools.getAllMoveActions((node.color + 1) % 2, map2);//移动指令变为对方可移动指令
                AttackActionList = mSpecTools.getAllAttackActions((node.color + 1) % 2, map2);//攻击指令变为对方可攻击指令
            }
            else//我方可用单位不为零
            {//下一手仍为我方行动
                nexTeamCol = node.color;//下一手仍然是自己的颜色
                ActionList = mSpecTools.getAllMoveActions(node.color, map2);//仍然是自己的移动行动指令
                AttackActionList = mSpecTools.getAllAttackActions(node.color, map2);//仍然是自己的攻击指令
            }

            ActionList.AddRange(AttackActionList);

            aTmp = aRand.Next(ActionList.Count);//随机数
            node.child.Add(new nodes());//新增子节点
            node.child[node.child.Count - 1].actions = ActionList[aTmp];//在最后加入攻击动作
            node.child[node.child.Count - 1].map = map2;//复制地图
            node.child[node.child.Count - 1].map.executeAction(ActionList[aTmp]);//执行攻击动作
            node.child[node.child.Count - 1].color = node.color;//子节点颜色为根节点颜色
            ActionList.RemoveAt(aTmp);//移去攻击指令
        }

        #endregion

        //此函数用于判断游戏的胜败
        private double evaluateMAP(Map map, int TeamColor)
        {

            int EnemyTeam = (TeamColor + 1) % 2;

            if (map.getTurnLimit() == map.getTurnCount())//如果超过了局数限制
            {
                if (map.getDrawHPThreshold() > Math.Abs(map.getNumOfAliveColorUnits(TeamColor) - map.getNumOfAliveColorUnits(EnemyTeam)))
                    return 0.5; //0.5值为和局
                else
                {
                    if (map.getNumOfAliveColorUnits(TeamColor) - map.getNumOfAliveColorUnits(EnemyTeam) > map.getDrawHPThreshold())
                        return 1;//赢局
                    else return 0;//输局
                }
            }
            //以下为没到和局
            if (map.getNumOfAliveColorUnits(Consts.BLUE_TEAM) == 0 && map.getNumOfAliveColorUnits(Consts.RED_TEAM) == 0) return 0;//同归于尽，输局
            if (map.getNumOfAliveColorUnits(EnemyTeam) == 0)
                return 1;//敌人死光，赢局
            if (map.getNumOfAliveColorUnits(TeamColor) == 0)
                return 0;//自己单位死光，输局

            return -1;//返回值为-1，即游戏没结束
        }


        //progressivewidening剪枝
        private void progressiveWiding(nodes rootnode, int i, List<Action> aAtAcList, List<Action> aMoAcList)
        {
            //根节点下子节点数目的设定，根据总模拟数进行确定
            int childnum = (int)(Math.Log(i / 40.0) / Math.Log(1.4) + 2);
            if (i >= 3000) childnum = (int)(Math.Log((i + 2000.0) / 45.0) / Math.Log(1.2) - 11);
            if (childnum < 1) childnum = 2;
            if (i == 1) childnum = 1;

            Map copyMap = new Map();//复制一份地图
            Random aRand = RandomT.GetThreadRandom();
            int aTmp;//临时变量
                     //计算和未来的子节点数目的差，增加子节点的数目
            if (rootnode.child.Count < childnum)
            {
                while (childnum != rootnode.child.Count)//计算值与根节点的子节点数不等时循环
                {
                    if (aAtAcList.Count != 0)//攻击指令不为零
                    {
                        aTmp = aRand.Next(aAtAcList.Count);//随机数                      
                        {
                            copyMap = rootnode.map.createDeepClone();//复制地图
                            rootnode.child.Add(new nodes());//新增子节点                       
                            rootnode.child[rootnode.child.Count - 1].actions = aAtAcList[aTmp];//在最后加入攻击动作
                            rootnode.child[rootnode.child.Count - 1].map = copyMap;//复制地图
                            rootnode.child[rootnode.child.Count - 1].map.executeAction(aAtAcList[aTmp]);//执行攻击动作
                            rootnode.child[rootnode.child.Count - 1].color = rootnode.color;//子节点颜色为根节点颜色
                            mSpecTools.storeSpec(rootnode.child[rootnode.child.Count - 1], aAtAcList[aTmp], rootnode.map);
                            aAtAcList.RemoveAt(aTmp);//移去攻击指令
                        }
                        if (filter(rootnode.child[rootnode.child.Count - 1]) == false)
                        {
                            rootnode.child.Remove(rootnode.child[rootnode.child.Count - 1]);
                            continue;
                        }                        
                    }
                    else//攻击指令执行完
                    {
                        if (aMoAcList.Count == 0) break;
                        aTmp = aRand.Next(aMoAcList.Count);
                        copyMap = rootnode.map.createDeepClone();
                        rootnode.child.Add(new nodes());                        
                        rootnode.child[rootnode.child.Count - 1].actions = aMoAcList[aTmp];
                        rootnode.child[rootnode.child.Count - 1].map = copyMap;
                        rootnode.child[rootnode.child.Count - 1].map.executeAction(aMoAcList[aTmp]);
                        rootnode.child[rootnode.child.Count - 1].color = rootnode.color;
                        mSpecTools.storemoveSpec(rootnode.child[rootnode.child.Count - 1], aMoAcList[aTmp], rootnode.map);
                        aMoAcList.RemoveAt(aTmp);//从aMoAcList中随机取出动作加入子节点，复制其全部信息进入rootnode
                    }
                }
            }
        }

        private bool filter(nodes node)//攻击禁忌在剪枝时实现
        {
            switch (node.opUnit_type) {
                case 0:if (node.targetUnit_type != 0 || node.targetUnit_type != 1)
                    {
                        return false;
                    }
                    else
                        return true;
                case 1:if (node.targetUnit_type == 4 || node.targetUnit_type == 0)
                    {
                        return false;
                    }
                    else
                        return true;
                case 2:
                        return true;
                case 3:
                        return true;
                case 4:
                        return true;
                case 5:
                        return true;
                default:
                        return true;
            }
        }

        private bool stopsuicide(nodes rootnode,nodes child)
        {
            Map map2 = new Map();
            map2 = rootnode.map;

            int[] damages = DamageCalculator.calculateDamages(map2, child.actions);//计算可能伤害
            int damageDif = damages[0] - damages[1];//计算伤害差

            if (damages[1] >= map2.getUnit(child.actions.operationUnitId).getHP() && damages[0] == 0)//反击大于自身HP
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        private Action MaxUnitdeterminate(nodes rootnode)//Sub AI attached in main&
        {
            Map map2 = new Map();
            map2 = rootnode.map;
            Action returnAction = new Action();
            Random aRand = RandomT.GetThreadRandom();
            int switcher = 0;
            int maxdamageID = -1;
            int maxdamage = 0;
            List<Action> attackAction = mSpecTools.getAllAttackActions(rootnode.color, map2);
            List<Action> defenceAction = mSpecTools.getAllMoveActions(rootnode.color, map2);

            Logger.addLogMessage("Activate SUB AI -Max Unit Determination Module" + "\r\n", rootnode.color);

            if (attackAction.Count != 0)
            {
                switcher = 1;
            }
            else
            {
                switcher = 0;
            }

            if (switcher == 1)//attack
            {
                for (int a = 0; a < attackAction.Count; a++)
                {
                    int[] damages = DamageCalculator.calculateDamages(map2, attackAction[a]);//计算可能伤害
                    int damageDif = damages[0] - damages[1];//计算伤害差
                    if (damages[0] > maxdamage)//伤害大于最大伤害
                    {
                        if (damages[0] >= map2.getUnit(attackAction[a].targetUnitId).getHP())//确保消灭
                        {
                            maxdamageID = a;
                            maxdamage = damages[0];
                            returnAction = attackAction[a];
                        }
                        else//无消灭可能
                        {
                            if (damageDif > maxdamage)
                            {
                                maxdamageID = a;
                                maxdamage = damageDif;
                                returnAction = attackAction[a];
                            }
                        }
                    }
                    else continue;
                }
            }
            if(maxdamage == 0)//伤害全为零 即无有效伤害
            {
                switcher = 0;
            }

            if (switcher == 0)//defence
            {
                List<Unit> myUnits = map2.getUnitsList(rootnode.color, false, true, false);
                List<Unit> enemyUnits = map2.getUnitsList(rootnode.color, false, false, true);
                Unit opUnit = myUnits[aRand.Next(myUnits.Count)];

                bool[,] movable = RangeController.getReachableCellsMatrix(opUnit, map2);

                int maxScore = Int32.MinValue;
                int maxX = -1;
                int maxY = -1;

                string matrixLog = "  ";

                for (int x = 1; x < map2.getXsize() - 1; x++)
                {
                    for (int y = 1; y < map2.getYsize() - 1; y++)
                    {
                        if (movable[x, y] == false)
                        {
                            matrixLog += " ---";
                            continue;
                        }

                        int defense = map2.getFieldDefensiveEffect(x, y);
                        if (opUnit.getSpec().isAirUnit() == true) defense = 0;

                        int localMaxScore = Int32.MinValue;
                        foreach (Unit enemyU in enemyUnits)
                        {
                            int dist = Math.Abs(x - enemyU.getXpos()) + Math.Abs(y - enemyU.getYpos());
                            int effect = opUnit.getSpec().getUnitAtkPower(enemyU.getSpec().getUnitType());

                            int score = defense * 5 + effect / (dist + 5);

                            if (score > maxScore)
                            {
                                maxScore = score;
                                maxX = x;
                                maxY = y;
                            }
                            if (score > localMaxScore) localMaxScore = score;
                        }
                        matrixLog += localMaxScore.ToString().PadLeft(4);
                    }
                    matrixLog += "\r\n  ";
                }
                return Action.createMoveOnlyAction(opUnit, maxX, maxY);
            }
            return returnAction;
        }

        #region 行动控制方法

        //Get the order of Units and store it in the List
        private void getorder(Map map, int teamColor, bool gameBegin)
        {
            if (gameBegin == true)
            {
                Map begin_map = map.createDeepClone();
                order = begin_map.getUnitsList(teamColor, false, true, false);
            }
        }

        private void doOneTurnAction(Map map, int teamColor)
        {
            while (doRandomAtkAction(map, teamColor)) ;
            while (doRandomMoveAction(map, teamColor)) ;
            map.resetActionFinish();
            map.incTurnCount();
        }

        private bool doRandomAtkAction(Map map, int teamColor)
        {
            List<Action> atkActions = AiTools.getAllAttackActions(teamColor, map);
            Random fRand = RandomT.GetThreadRandom();
            if (atkActions.Count == 0) return false;
            int tmp = fRand.Next(atkActions.Count);
            map.executeAction(atkActions[tmp]);
            return true;
        }

        private bool doRandomMoveAction(Map map, int teamColor)
        {
            List<Action> moveActions = AiTools.getAllMoveActions(teamColor, map);
            Random fRand = RandomT.GetThreadRandom();
            if (moveActions.Count == 0) return false;
            int tmp = fRand.Next(moveActions.Count);
            map.executeAction(moveActions[tmp]);
            return true;
        }

        public static class RandomT
        {
            private static int seed = Environment.TickCount;

            private static ThreadLocal<Random> randomWrapper = new ThreadLocal<Random>(() =>
                new Random(Interlocked.Increment(ref seed))
            );

            public static Random GetThreadRandom()
            {
                return randomWrapper.Value;
            }
        }

        #endregion
    }
}

